#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################

def doraise(obj, **kwargs):
    return bool(kwargs.pop('raiseError', getattr(obj, 'raiseError', True)))

def raiseIt(exc, raiseoupas, *args, **kwargs):
    exc = exc(*args, **kwargs)
    if raiseoupas:
        raise exc
    else:
        print("\033[31m"+exc.message+"\033[39m")
        return True
    return False

class OIException(Exception):
    """
    Root for SOIF Exceptions, only used to trigger any soif errors, never raised
    """
    def __init__(self, *args, **kwargs):
        self.args = [a for a in args] + [a for a in kwargs.values()]
    def __str__(self):
        return repr(self.message)
    def __repr__(self):
        return repr(self.message)

class NoTargetTable(OIException):
    """
    If the file has no OITARGET table
    """
    def __init__(self, src="", *args, **kwargs):
        super(NoTargetTable, self).__init__(src, *args, **kwargs)
        self.message = "There's no OI_TARGET table in the oifits file '%s'! Go get some coffee!" % (src)

class NoWavelengthTable(OIException):
    """
    If the file has no OITARGET table
    """
    def __init__(self, src="", *args, **kwargs):
        super(NoWavelengthTable, self).__init__(src, *args, **kwargs)
        self.message = "There's no OI_WAVELENGTH table in the oifits file '%s'! You're pretty much screwed!" % (src)

class ReadOnly(OIException):
    """
    If the parameter is read-only
    """
    def __init__(self, attr, *args, **kwargs):
        super(ReadOnly, self).__init__(attr, *args, **kwargs)
        self.message = "Attribute '%s' is read-only" % (attr)

class InvalidDataType(OIException):
    """
    If the data type provided does not exist
    """
    def __init__(self, datatype, *args, **kwargs):
        super(InvalidDataType, self).__init__(datatype, *args, **kwargs)
        self.message = "Data type '%s' does not exist" % (datatype)

class HduDatatypeMismatch(OIException):
    """
    If the data type and the hdu provided do not match
    """
    def __init__(self, hduhead, datatype, *args, **kwargs):
        super(HduDatatypeMismatch, self).__init__(hduhead, datatype, *args, **kwargs)
        self.message = "Data type '%s' and hdu with '%s' data do not match" % (datatype, hduhead)

class BadMaskShape(OIException):
    """
    If the mask shape does not match the data shape
    """
    def __init__(self, shape, *args, **kwargs):
        super(BadMaskShape, self).__init__(shape, *args, **kwargs)
        self.message = "Bad mask shape. Should be '%s'" % shape

class WrongData(OIException):
    """
    If the data provided has the wrong data type
    """
    def __init__(self, typ, *args, **kwargs):
        super(WrongData, self).__init__(typ, *args, **kwargs)
        self.message = "Wrong data given, should be '%s'" % typ

class IncompatibleData(OIException):
    """
    If the data type and the hdu provided do not match
    """
    def __init__(self, typ1, typ2, *args, **kwargs):
        super(IncompatibleData, self).__init__(typ1, typ2, *args, **kwargs)
        self.message = "Can't merge '%s' and '%s'" % (typ1, typ2)

class NotADataHdu(OIException):
    """
    If the hdu provided does not contain data
    """
    def __init__(self, idx, src, *args, **kwargs):
        super(NotADataHdu, self).__init__(idx, src, *args, **kwargs)
        self.message = "Hdu index '%s' in file '%s' does not contain data" % (idx, src)

class NoSystematicsFit(OIException):
    """
    If the user did not set on the fit of systematics
    """
    def __init__(self, *args, **kwargs):
        super(NoSystematicsFit, self).__init__(*args, **kwargs)
        self.message = "You are not fitting systematics"

class NotCallable(OIException):
    """
    If the function is callable
    """
    def __init__(self, fct, *args, **kwargs):
        super(NotCallable, self).__init__(fct, *args, **kwargs)
        self.message = "'%s' should be callable" % fct



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################


#import matplotlib.pyplot as _plt
from time import strftime as _strftime
try:
    import astropy.io.fits as _pf
except ImportError:
    import pyfits as _pf

import oiexception as _exc
import _core
_np = _core.np


class Oigrab(object):
    """
    Opens, reads and filters data contained in the OIFITS file ``src``.

    Args:
      * src (str): the path (relative or absolute) to the OIFITS file
    
    Kwargs:
      * raiseError (bool): if ``True``, raises errors, otherwise prints them

    Raises:
      * NoTargetTable: if the OIFITS file has no OI_TARGET table

    >>> import soif.oidata as oidata
    >>> data = oidata.Oigrab('./data/datafile.oifits')
    """
    def __init__(self, src, **kwargs):
        self._init(src=src, **kwargs)

    def _init(self, src, **kwargs):
        self.src = str(src)
        self.raiseError = bool(kwargs.pop('raiseError', True))
        hdus = _pf.open(self.src)
        for item in hdus:
            if item.header.get('EXTNAME')=='OI_TARGET':
                hdutgt = item
                break
        else:
            hdus.close()
            if _exc.raiseIt(_exc.NoTargetTable, self.raiseError, src=self.src): return False
        self._targets = {}
        for ind, tgt in zip(hdutgt.data["TARGET_ID"], hdutgt.data["TARGET"]):
            self._targets[ind] = tgt
        hdus.close()

    def _info(self):
        return "%s<SOIF File>%s\n File: '%s'" % (_core.font.blue, _core.font.normal, self.src)
    def __repr__(self):
        return self._info()
    def __str__(self):
        return self._info()

    @property
    def targets(self):
        return self._targets
    @targets.setter
    def targets(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="targets")

    def show_specs(self, ret=False, **kwargs):
        """
        Gets the target list and the data details from the OIFITS file.

        Args:
          * ret (bool): if ``True``, returns the information, otherwise prints it

        Returns:
          * a dictionary {'hdu index:info'} where info corresponds to a list of (Acquisition index, Target ID, MJD, N(UV), N(wl)) tuples

        >>> import soif.oidata as oidata
        >>> data = oidata.Oigrab('./data/datafile.oifits')
        >>> data.showspecs()
        TARGETS:
        1: IRS_48
        2: Elia_2-15

        VIS2 [hdu=3]:
        Acq. Index | Target ID |      MJD       |  UVs | N wl
        -----------------------------------------------------
                 0 |         1 | 55636.3827746  |   21 |    1
                 1 |         2 | 55636.3827989  |   21 |    1
                 2 |         1 | 55636.3828232  |   21 |    1
        """
        hdus = _pf.open(self.src)
        if not ret:
            print("TARGETS:")
            for ind, tgt in self.targets.items():
                print('%d: %s' % (ind, tgt))
        tgtlist = {}
        for idx, item in enumerate(hdus):
            if _core.hduToDataType(item) is not None:
                targetindexnumber, MJD, (ndata, nset, nunique, nholes, nwl) = _core.gethduMJD(item, withdet=True)
                tgtlist[idx] = []
                if not ret: print("\n%s [hdu=%d]:\nAcq. Index | Target ID |      MJD      |  UVs | N wl\n%s" % (_core.hduToDataType(item), idx, "-"*52))
                for tgtidx, sMJD in zip(targetindexnumber.reshape((-1, nunique))[:,0], MJD[targetindexnumber].reshape((-1, nunique))[:,0]):
                    tgtfilter = slice(tgtidx*nunique, (tgtidx+1)*nunique)
                    tgtid = item.data['TARGET_ID'][tgtfilter]
                    if not ret: print("%10s | %9s | %13s | %4s | %4s"%(tgtidx, tgtid[0], sMJD, tgtid.size, nwl))
                    tgtlist[idx].append((tgtidx, tgtid[0], sMJD, tgtid.size, nwl))
        hdus.close()
        if ret: return tgtlist

    def show_filtered(self, tgt=None, mjd=[None, None], hduNums=[], vis2=True, t3phi=True, t3amp=True, visphi=True, visamp=True, verbose=False, **kwargs):
        """
        Given an oifits file 'src' and filtering parameters on the target name (OI_TARGET table), the instrument name (OI_WAVELENGTH table), the array name (OI_ARRAY table), the observation wavelength (OI_WAVELENGTH table) and the acquisition time [t_min, t_max] (OI_VIS2, OI_VIS, OI_T3 tabkes), this function returns the data indices of the data matching all of these different filters.
        These lists are used to load the data within an Oidata object.

        Leave input parameter to 'None' to discard filtering on that particular parameter.

        Returns: VIS2, T3, VIS indeces as a tuple of 3 lists
        """
        hdus = _pf.open(self.src)
        mjd = [float(mjd[0] if mjd[0] is not None else -_np.inf), float(mjd[1] if mjd[1] is not None else _np.inf)]
        datayouwant = {'data':{'VIS2':bool(vis2), 'T3PHI':bool(t3phi), 'T3AMP':bool(t3amp), 'VISPHI':bool(visphi), 'VISAMP':bool(visamp)}}
        for idx, item in enumerate(hdus):
            if _core.aslist(hduNums) != [] and idx not in _core.aslist(hduNums): continue
            if _core.hduToDataType(item) is not None:
                MJD = _core.gethduMJD(item)[1]
                filt = ((MJD>=mjd[0]) & (MJD<=mjd[1]))
                if tgt is not None:
                    filt = (filt & (item.data.field("TARGET_ID") == int(tgt)))
                if verbose: print("%s:\n  %d/%s\n"%(_core.hduToDataType(item), filt.sum(), item.data["TARGET_ID"].size))
                datayouwant[idx] = _np.arange(item.data["TARGET_ID"].size)[filt]
        hdus.close()
        return datayouwant

    def extract(self, tgt=None, mjd=[None, None], wl=[None, None], hduNums=[], vis2=True, t3phi=True, t3amp=True, visphi=True, visamp=True, flatten=False, degree=True, significant_figures=5, erb_sigma=None, sigma_erb=None, systematic_prior=None, systematic_bounds=None, verbose=False, **kwargs):
        datayouwant = self.show_filtered(tgt=tgt, mjd=mjd, vis2=vis2, hduNums=hduNums, t3phi=t3phi, t3amp=t3amp, visphi=visphi, visamp=visamp, verbose=verbose, **kwargs)
        return Oifits(src=self.src, datafilter=datayouwant, flatten=flatten, degree=degree, significant_figures=significant_figures, wl=wl, erb_sigma=erb_sigma, sigma_erb=sigma_erb, systematic_prior=systematic_prior, systematic_bounds=systematic_bounds, **kwargs)


class OidataEmpty(object):
    def __init__(self, datatype, **kwargs):
        self.raiseError = bool(kwargs.pop('raiseError', True))
        self.datatype = str(datatype).upper()
        if self.datatype not in _core.ATTRDATATYPE.keys():
            if _exc.raiseIt(_exc.InvalidDataType, self.raiseError, datatype=self.datatype): return False
        self._has = False
        self._useit = False

    def _info(self):
        return "%s data, None" % (self.datatype)
    def __repr__(self):
        return self._info()
    def __str__(self):
        return self._info()

    @property
    def useit(self):
        return (self._useit and self._has)
    @useit.setter
    def useit(self, value):
        self._useit = bool(value)
    
    def __bool__(self):
        return self._has
    def __nonzero__(self):
        return self.__bool__()

    def __getitem__(self, key):
        return getattr(self, "_"+key)


class Oidata(OidataEmpty):
    def __init__(self, src, hduidx, datatype, hduwlidx, indices=[], wlindices=[], degree=True, flatten=False, significant_figures=5, **kwargs):
        super(Oidata, self).__init__(datatype=datatype, **kwargs)
        self._input_src = [str(src)]
        self._input_hduidx = [int(hduidx)]
        self._input_hduwlidx = [int(hduwlidx)]
        hdus = _pf.open(self._input_src[-1])
        hdu = hdus[self._input_hduidx[-1]]
        hduwl = hdus[self._input_hduwlidx[-1]]

        if self.datatype not in _core.ATTRDATATYPE.keys():
            if _exc.raiseIt(_exc.InvalidDataType, self.raiseError, datatype=self.datatype): return False
        if _core.DATAKEYSDATATYPE[self.datatype]['data'] not in _core.hduToColNames(hdu):
            if _exc.raiseIt(_exc.HduDatatypeMismatch, self.raiseError, hduhead=_core.hduToDataType(hdu), datatype=self.datatype): return False

        self._input_degree = [bool(degree)]
        self._input_flatten = [bool(flatten)]
        self._input_significant_figures = [int(significant_figures)]
        self._input_indices = [list(indices)]
        self._input_wlindices = [list(wlindices)]
        
        self._has = True
        self._useit = True
        wlindices = slice(0, 10000000) if wlindices == [] else _np.asarray(wlindices).ravel()
        indices = slice(0, 10000000) if indices == [] else _np.asarray(indices).ravel()

        # attributes to be copy-pasted
        for key, vl in _core.ATTRDATATYPE[self.datatype].items():
            setattr(self, '_'+key, vl)

        # wavelength attributes to be extracted out of OIFITS
        for key, vl in _core.KEYSWL.items():
            setattr(self, '_'+key, hduwl.data[vl].ravel()[wlindices])
        self._wlsize = self._wl.size
        self._datasize = hdu.data[_core.DATAKEYSDATATYPE[self.datatype]['data']][indices].shape[0]

        # data attributes to be extracted out of OIFITS
        for key, vl in _core.DATAKEYSDATATYPE[self.datatype].items():
            setattr(self, '_'+key, hdu.data[vl][indices].reshape((self._datasize, -1))[:,wlindices])
        self.mask = _np.logical_not(self._mask) # swaps Trues to Falses to have a good mask

        # data attributes to be extracted out of OIFITS and replicated with wl
        for key, vl in _core.UVKEYSDATATYPE[self.datatype].items():
            setattr(self, '_'+key, _core.replicate(hdu.data[vl][indices].ravel(), (None, self._wlsize)))

        # done with the file
        hdus.close()

        # add the data dimension on wl-like attributes
        for key in _core.KEYSWL.keys():
            setattr(self, "_"+key, _core.replicate(getattr(self, "_"+key), (self._datasize, None)))

        # combine the UV coordinates of T3
        if self.is_t3:
            self._u = _np.concatenate((_np.expand_dims(self._u1, -1), _np.expand_dims(self._u2, -1), _np.expand_dims(-self._u1-self._u2, -1)), axis=-1)
            self._v = _np.concatenate((_np.expand_dims(self._v1, -1), _np.expand_dims(self._v2, -1), _np.expand_dims(-self._v1-self._v2, -1)), axis=-1)
            # delete temporary attributes
            for key in _core.UVKEYSDATATYPE[self.datatype].keys():
                delattr(self, "_"+key)
            # add the data dimension on wl-like attributes
            for key in _core.KEYSWL.keys():
                setattr(self, "_"+key, _core.replicate(getattr(self, "_"+key), (None, 3)))

        # convert to radian if needed
        if self.is_angle and degree:
            self._data = self._data*_core.DEG2RAD
            self._error = self._error*_core.DEG2RAD

        self.significant_figures = min(8, max(1, int(significant_figures)))
        self._flat = bool(flatten)
        if self._flat:
            self.flatten()
        else:
            self.update()

    def _info(self):
        return "%s data, shape: %s, wl: %s%s" % (self.datatype, _core.maskedshape(self.shapedata, _np.logical_not(self.mask).sum()), self._wlmin, (" to "+str(self._wlmax))*int(self._wlspan!=0))
    
    @property
    def data(self):
        if self._use_mask:
            return self._data[self.mask]
        else:
            return self._data
    @data.setter
    def data(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="data")

    @property
    def error(self):
        if self._use_mask:
            return self._error[self.mask]
        else:
            return self._error
    @error.setter
    def error(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="error")

    @property
    def mask(self):
        return self._mask
    @mask.setter
    def mask(self, value):
        if value is True or value is False or value is None: # removes the mask
            self._use_mask = False
            self._mask = np.ones(self._data.shape, dtype=bool)
        else:
            value = _np.asarray(value, dtype=bool)
            if value.shape != self._data.shape:
                if _exc.raiseIt(_exc.BadMaskShape, self.raiseError, shape=str(self._data.shape)): return False
            self._use_mask = not value.all()
            self._mask = value

    @property
    def u(self):
        if self._use_mask:
            return self._u[self.mask]
        else:
            return self._u
    @u.setter
    def u(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="u")

    @property
    def v(self):
        if self._use_mask:
            return self._v[self.mask]
        else:
            return self._v
    @v.setter
    def v(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="v")

    @property
    def wl(self):
        if self._use_mask:
            return self._wl[self.mask]
        else:
            return self._wl
    @wl.setter
    def wl(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="wl")

    @property
    def wl_d(self):
        if self._use_mask:
            return self._wl_d[self.mask]
        else:
            return self._wl_d
    @wl_d.setter
    def wl_d(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="wl_d")

    @property
    def shapedata(self):
        return self.data.shape
    @shapedata.setter
    def shapedata(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="shapedata")

    @property
    def shapeuv(self):
        return self.u.shape
    @shapeuv.setter
    def shapeuv(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="shapeuv")

    @property
    def is_angle(self):
        return self._is_angle
    @is_angle.setter
    def is_angle(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="is_angle")

    @property
    def is_t3(self):
        return self._is_t3
    @is_t3.setter
    def is_t3(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="is_t3")

    @property
    def flat(self):
        return self._flat
    @flat.setter
    def flat(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="flat")

    def flatten(self, **kwargs):
        if self.is_t3:
            for key in _core.KEYSUV:
                setattr(self, "_"+key, getattr(self, "_"+key).reshape((-1, 3)))
            for key in _core.KEYSDATA:
                setattr(self, "_"+key, getattr(self, "_"+key).ravel())
        else:
            for key in _core.KEYSDATA + _core.KEYSUV:
                setattr(self, "_"+key, getattr(self, "_"+key).ravel())
        self.update()
        self._flat = True

    def _addData(self, data, flatten=True, **kwargs):
        if not isinstance(data, OidataEmpty): return # trivial, nothing to add
        if not isinstance(data, Oidata):
            if _exc.raiseIt(_exc.WrongData, _exc.doraise(self, **kwargs), typ='Oidata'): return False
        if self.datatype != data.datatype:
            if _exc.raiseIt(_exc.IncompatibleData, _exc.doraise(self, **kwargs), typ1=self.datatype, typ2=data.datatype): return False
        # do we flatten it?
        if flatten or self.shapedata[-1] != data.shapedata[-1]:
            self.flatten()
            data.flatten()
        # concatenate data
        for key in _core.KEYSDATA + _core.KEYSUV:
            setattr(self, "_"+key, _np.concatenate((getattr(self, "_"+key), getattr(data, "_"+key)), axis=0))
        # update input keys
        for key in _core.INPUTSAVEKEY:
            setattr(self, "_input_"+key, getattr(self, "_input_"+key) + getattr(data, "_input_"+key))
        # update the data
        self.update()

    def update(self):
        """
        Given u, v, wl and flag information as object properties, this function updates the Oidata object: the data masking (from the new mask property) and the bl, pa, blwl properties (from u, v and wl properties)
        """
        self._u = _core.round_fig(self._u, self.significant_figures)
        self._v = _core.round_fig(self._v, self.significant_figures)
        self._wl = _core.round_fig(self._wl, self.significant_figures)
        self._wl_d = _core.round_fig(self._wl_d, self.significant_figures)
        self._invvar = 1./self.error**2
        self.bl = _core.round_fig(_np.hypot(self.v, self.u), self.significant_figures)
        self.pa = _core.round_fig(_np.arctan2(self.v, self.u), self.significant_figures)
        self.blwl = _core.round_fig(self.bl/self.wl, self.significant_figures)
        self._wlmin = self.wl.min()
        self._wlmax = self.wl.max()
        self._wlspan = self._wlmax - self._wlmin



class Oifits(object):
    """
    This class opens, reads and sorts data contained in the file 'src' (oifits format).

    vis2, vis and t3 are lists of which corresponding data indeces to extract from the file. 
    """
    def __init__(self, src, datafilter, wl=[None, None], erb_sigma=None, sigma_erb=None, systematic_prior=None, systematic_bounds=None, flatten=False, degree=True, significant_figures=5, **kwargs):
        self.raiseError = bool(kwargs.pop('raiseError', True))
        # initialize empty data
        for key in _core.DATAKEYSLOWER:
            setattr(self, key, OidataEmpty(key))

        self.erb_sigma = _core.ident if erb_sigma is None else erb_sigma
        if not callable(self.erb_sigma):
            if _exc.raiseIt(_exc.NotCallable, self.raiseError, fct="erb_sigma"): return False
        self.sigma_erb = _core.ident if sigma_erb is None else sigma_erb
        if not callable(self.sigma_erb):
            if _exc.raiseIt(_exc.NotCallable, self.raiseError, fct="sigma_erb"): return False
        self.systematic_bounds = None if systematic_bounds is None else list(map(float, list(systematic_bounds)[:2]))
        self.systematic_prior = None if systematic_prior is None else float(systematic_prior)
        self._systematic_prior = self.systematic_prior
        # init parameters
        self.addData(src=str(src), datafilter=dict(datafilter), flatten=bool(flatten), degree=bool(degree), significant_figures=int(significant_figures), wl=list(wl)[:2], noupdate=True, **kwargs)
        self.update()

    def _info(self):
        txt = "\n".join([" "+str(getattr(self, key)) for key in _core.DATAKEYSLOWER if getattr(self, key).useit])
        if txt == "": txt = " No data"
        txt = "%s<SOIF Data>%s\n%s" % (_core.font.blue, _core.font.normal, txt)
        return txt
    def __repr__(self):
        return self._info()
    def __str__(self):
        return self._info()

    def addData(self, src, datafilter={}, flatten=False, degree=True, significant_figures=5, wl=[None, None], **kwargs):
        if isinstance(src, Oifits):
            thedata = src
            for datatype in _core.DATAKEYSLOWER:
                if not getattr(thedata, datatype): continue
                if getattr(self, datatype):
                    getattr(self, datatype)._addData(getattr(thedata, datatype), flatten=flatten, **kwargs)
                else:
                    setattr(self, datatype, getattr(thedata, datatype))
        else:
            whichdata = datafilter.get('data', {'data':{'VIS2':True, 'T3PHI':True, 'T3AMP':True, 'VISPHI':True, 'VISAMP':True}})
            hdus = _pf.open(src) # open
            hduwlidx = _core.hduWlindex(hdus)
            if not hduwlidx:
                if _exc.raiseIt(_exc.NoWavelengthTable, self.raiseError, src=src): return
            # get wl sorted
            wl = [float(wl[0] if wl[0] is not None else -_np.inf), float(wl[1] if wl[1] is not None else _np.inf)]
            allwl = hdus[hduwlidx].data[_core.KEYSWL['wl']]
            wlindices = _np.arange(allwl.size)[((allwl>=wl[0]) & (allwl<wl[1]))]
            # for each datafilter
            for idx, indices in datafilter.items():
                if not isinstance(idx, int) or _np.size(indices)==0: continue
                # if real data
                if _core.hduToDataType(hdus[idx]) is not None:
                    for datatype in _core.ALLDATAEXTNAMES[hdus[idx].header['EXTNAME']]:
                        if not whichdata.get(datatype.upper(), True): continue
                        thedata = Oidata(src=src, hduidx=idx, datatype=datatype, hduwlidx=hduwlidx, indices=indices, wlindices=wlindices, degree=degree, flatten=flatten, significant_figures=significant_figures, **kwargs)
                        if getattr(self, datatype.lower()):
                            getattr(self, datatype.lower())._addData(thedata, flatten=flatten, **kwargs)
                        else:
                            setattr(self, datatype.lower(), thedata)
                else:
                    hdus.close()
                    if _exc.raiseIt(_exc.NotADataHdu, self.raiseError, idx=idx, src=str(src)): return
            hdus.close()
        if not kwargs.pop('noupdate', False): self.update()

    def flatten(self):
        """
        Flattens all data contained in the Oidata object. This can be useful in order to add several bits of data that do not have the same shapes
        """
        for key in _core.DATAKEYSLOWER:
            if getattr(self, key): getattr(self, key).flatten()

    def update(self):
        """
        Updates all data contained in the Oidata object
        """
        for key in _core.DATAKEYSLOWER:
            if getattr(self, key): getattr(self, key).update()

        funkydtypeint = [('u', int), ('v', int), ('wl', int)]
        funkydtype = [('u', _np.float32), ('v', _np.float32), ('wl', _np.float32)]

        # get uvwl sets as integer to extract uniques
        unique_uvwl = _np.zeros(0, dtype=funkydtypeint)
        for key in _core.DATAKEYSLOWER:
            thedata = getattr(self, key)
            if thedata:
                dum = _np.zeros(thedata.shapeuv, dtype=funkydtypeint)
                dum['u'] = _core.round_fig(x=thedata.u, n=thedata.significant_figures, retint=True)
                dum['v'] = _core.round_fig(x=thedata.v, n=thedata.significant_figures, retint=True)
                dum['wl'] = _core.round_fig(x=thedata.wl, n=thedata.significant_figures, retint=True)
                # deal with symmetry
                inv = (dum['u']<0)
                dum['u'][inv] *= -1
                dum['v'][inv] *= -1
                # stack
                unique_uvwl = _np.hstack((unique_uvwl, dum.flatten()))
        if _core.OLDNUMPY:
            uvwlind = _np.unique((unique_uvwl['u']+1j*unique_uvwl['v'])/unique_uvwl['wl'], return_index=True)[1]
        else:
            uvwlind = _np.unique(unique_uvwl, return_index=True)[1]
        # get uvwl sets as floats for calculations
        unique_uvwl = _np.zeros(0, dtype=funkydtype)
        for key in _core.DATAKEYSLOWER:
            thedata = getattr(self, key)
            if thedata:
                dum = _np.zeros(thedata.shapeuv, dtype=funkydtype)
                dum['u'] = _core.round_fig(x=thedata.u, n=thedata.significant_figures)
                dum['v'] = _core.round_fig(x=thedata.v, n=thedata.significant_figures)
                dum['wl'] = _core.round_fig(x=thedata.wl, n=thedata.significant_figures)
                # simplify symmetry
                inv = (dum['u']<0)
                dum['u'][inv] *= -1
                dum['v'][inv] *= -1
                # save the phase symmetry
                if thedata.is_angle:
                    thedata._phasesign = 1-inv*2
                # stack
                unique_uvwl = _np.hstack((unique_uvwl, dum.flatten()))
                thedata._ind = dum.copy()
        # extract uniques
        dum = _np.zeros(uvwlind.shape, dtype=funkydtype)
        dum['u'], dum['v'], dum['wl'] = unique_uvwl['u'][uvwlind], unique_uvwl['v'][uvwlind], unique_uvwl['wl'][uvwlind]
        unique_uvwl = dum
        # save indices
        for key in _core.DATAKEYSLOWER:
            thedata = getattr(self, key)
            if thedata:
                dum = _np.zeros(thedata._ind.shape, dtype='int')
                for i, v in enumerate(unique_uvwl):
                    dum[thedata._ind == v] = i
                thedata._ind = dum.copy()
        # prepare pre-processed uniques
        self.uvwl = {'u':unique_uvwl['u'], 'v':unique_uvwl['v'], 'wl':unique_uvwl['wl'], 'blwl':_np.hypot(unique_uvwl['u'], unique_uvwl['v'])/unique_uvwl['wl']}
        self._wlmin = self.uvwl['wl'].min()
        self._wlmax = self.uvwl['wl'].max()
        self._wlspan = self._wlmax - self._wlmin

    @property
    def systematic_fit(self):
        return isinstance(self.systematic_bounds, list)
    @systematic_fit.setter
    def systematic_fit(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="systematic_fit")

    def systematic_p0(self):
        if self.systematic_fit:
            randomizer = _core.gen_generator()
            return randomizer.uniform(low=self.systematic_bounds[0], high=self.systematic_bounds[1])
        else:
            _exc.raiseIt(NoSystematicsFit, self.raiseError)

    def remorph(self, viscomp):
        ret = []
        for key in _core.DATAKEYSLOWER:
            thedata = getattr(self, key)
            if thedata:
                retdbl = isinstance(viscomp, (tuple, list))
                if retdbl:
                    viscomp, flx = viscomp
                if thedata.is_t3:
                    if thedata.is_angle: # t3phi
                        dum = (_core.FCTVISCOMP[key.upper()](viscomp)[thedata._ind]*thedata._phasesign).sum(-1)
                    else: # t3amp
                        dum = (_core.FCTVISCOMP[key.upper()](viscomp[thedata._ind])).prod(-1)
                else:
                    dum = _core.FCTVISCOMP[key.upper()](viscomp)[thedata._ind]
                    if thedata.is_angle: # visphi
                        dum *= thedata._phasesign
                if retdbl:
                    ret.append((dum, flx[thedata._ind]))
                else:
                    ret.append(dum)
            else:
                ret.append(None)
        return ret


    def save(self, filename, append=False, clobber=False):
        ext = '.oif.fits'
        if filename.find(ext)==-1: filename += ext
        if append:
            hdulist = _pf.open(filename, mode='append')
        else:
            hdulist = _pf.HDUList()
        allmodes = ['vis2', 't3phi', 't3amp', 'visphi', 'visamp']
        for i in range(len(self._input_src)):
            for mode in allmodes:
                if getattr(self,"_has"+mode):
                    hdu = _pf.PrimaryHDU()
                    hdu.header.set('EXT', 'DATA', comment='Type of information in the HDU')
                    hdu.header.set('DATE', _strftime('%Y%m%dT%H%M%S'), comment='Creation Date')
                    hdu.header.set('DATAFILE', i, comment='Data file number')
                    hdu.header.set('SRC', str(self._input_src[i]), comment='Path to the datafile')
                    hdu.header.set('DATATYPE', mode, comment='Type of data from datafile '+str(i))
                    hdu.header.set('FLATTEN', bool(self._input_flatten[i]), comment='Should the data be flatten')
                    hdu.header.set('SIG_FIG', int(self.significant_figures), comment='Significant figures for u,v coordinates')
                    if mode in ['visphi', 't3phi']:
                        hdu.header.set('DEGREES', bool(self._input_degrees[i]), comment='Is datafile in degrees')
                    hdu.data = _np.ravel(getattr(self, "_input_"+mode)[i])

                    hdu.header.add_comment('Measurement indices to be imported from datafile for the given datatype.')
                    hdu.header.add_comment('Written by Guillaume SCHWORER')
                    hdulist.append(hdu)
        for mode in allmodes:
            if getattr(self,"_has"+mode):
                hdu = _pf.PrimaryHDU()
                hdu.header.set('EXT', 'DATAMASK', comment='Type of information in the HDU')
                hdu.header.set('DATE', _strftime('%Y%m%dT%H%M%S'), comment='Creation Date')
                hdu.header.set('DATATYPE', mode, comment='Type of data from datafile '+str(i))
                hdu.data = getattr(self, mode).mask.astype(_np.uint8)
                hdu.header.add_comment('Data mask for the given datatype for all data contained in the different files.')
                hdu.header.add_comment('Written by Guillaume SCHWORER')
                hdulist.append(hdu)

        if append:
            hdulist.flush()
            hdulist.close()
        else:
            hdulist.writeto(filename, clobber=clobber)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################


import matplotlib.pyplot as _plt
from matplotlib.gridspec import GridSpec as _matplotlibGridspecGridSpec
from matplotlib.patches import FancyArrowPatch as _matplotlibPatchesFancyArrowPatch
from oiunitmodels import _objects
from _oiunitmodels import *
from oimainobject import Oimainobject as _Oimainobject
from copy import deepcopy as _deepcopy
from time import strftime as _strftime
try:
    import astropy.io.fits as _pf
except ImportError:
    import pyfits as _pf

import oiexception as _exc
import _core
_np = _core.np


class Oimodel(object):
    def __init__(self, oidata, objs=[], tweakparams=None, **kwargs):
        self._objs = []
        self.raiseError = bool(kwargs.pop('raiseError', True))
        self.oidata = oidata
        if tweakparams is not None and not callable(tweakparams):
            if _exc.raiseIt(_exc.NotCallable, self.raiseError, fct="tweakparams"): return False
        self._tweakparams = tweakparams
        self.nobj = 0
        self._nparamsObj = 0
        if not hasattr(objs, "__iter__"):
            self.add_obj(objs)
        else:
            for item in objs:
                self.add_obj(item)

    def _info(self):
        return _core.font.blue+"<SOIF Model>%s\n %s objects:\n  %s\n%s"%(_core.font.normal, self.nobj, "\n  ".join(map(str, self._objs)), str(self.oidata))
    def __repr__(self):
        return self._info()
    def __str__(self):
        return self._info()


    @property
    def nparams(self):
        return self._nparamsObj + self.oidata.systematic_fit
    @nparams.setter
    def nparams(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="nparams")

    @property
    def nparamsObjs(self):
        self._nparamsObj = 0
        for item in self._objs: self._nparamsObj += item._nparams
        return self._nparamsObj
    @nparamsObjs.setter
    def nparamsObjs(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="nparamsObjs")
    

    def add_obj(self, typ, name=None, params={}, prior={}):
        """
        Add an object to the model
        """
        # gets the future new name of the object
        if isinstance(typ, _Oimainobject):
            name = typ.name
        else:
            name = _core.clean_name(name)
        # check for already existing name
        if hasattr(self, "o_"+name): raise NameError("Object name already exists")
        # if all good, proceeds
        if isinstance(typ, _Oimainobject): # if first argument is already a built object
            self._objs.append(_deepcopy(typ))
        elif typ in _objects: # if we need to build the object, and it exists
            self._objs.append(globals()[typ](name=name, params=params, prior=prior))
        else:
            print(_core.font.red+"ERROR: Could not find the object name for '%s', name given: %s%s" % (typ, name, _core.font.normal))
            return
        setattr(self, "o_"+name, self._objs[-1]) # quick access as a class property
        if hasattr(self._objs[-1], '_prepare'): self._objs[-1]._prepare(oidata=self.oidata)
        self.nobj += 1
        dum = self.nparamsObjs
    
    def del_obj(self, idobj):
        """
        Delete an object from the model.
        idobj can be the name of the object or its index in the model list 
        """
        if not isinstance(idobj, int):
            name = _core.clean_name(name)
            if not hasattr(self, "o_"+name): raise NameError("Didn't find object name")
            ind = self._objs.index(getattr(self, "o_"+name))
        else:
            ind = name
            name = self._objs[ind].name
        dummy = self._objs.pop(ind)
        delattr(self, "o_"+name)
        print(_core.font.blue+"Deleted object '%s'." % (name)+_core.font.normal)
        self.nobj -= 1
        dum = self.nparamsObjs

    def getP0(self):
        """
        Return an initialized param list
        """
        ret = []
        for item in self._objs:
            ret += item.getP0()
        if self.oidata.systematic_fit: ret += [self.oidata.systematic_p0()]
        return ret


    @property
    def paramstr(self):
        ret = []
        for item in self._objs:
            for arg in item._pkeys:
                ret.append(item.name+"_"+arg)
        if self.oidata.systematic_fit: ret += ["sys"]
        return ret
    @paramstr.setter
    def paramstr(self, value):
        _exc.raiseIt(_exc.ReadOnly, self.raiseError, attr="paramstr")


    @property
    def params(self):
        """
        Return current params
        """
        ret = []
        for item in self._objs:
            ret += getattr(item, "params", [])
        if self.oidata.systematic_fit: ret += [self.oidata.systematic_prior if self.oidata.systematic_prior is not None else self.oidata.systematic_p0()]
        return ret
    @params.setter
    def params(self, value):
        self.setParams(value)


    def reinit(self, params):
        self.setParams(params=params, priors=True)


    def statut(self, params, customlike=None, **kwargs):
        self.setParams(params)
        chi2 = self.likelihood(params=params, customlike=customlike, chi2=True, **kwargs)
        print('Khi-2: %.3f' % chi2)
        print('Khi-2 NULL: %.3f' % self.likelihood(params=params, customlike=customlike, chi2=True, null=True, **kwargs))
        print('ln-like: %.3f' % self.likelihood(params=params, customlike=customlike, chi2=False, **kwargs))
        print('ln-like NULL: %.3f' % self.likelihood(params=params, customlike=customlike, chi2=False, null=True, **kwargs))
        print('Number of parameters (k): %i' % self._nparamsObj)


    def setParams(self, params, priors=False):
        if len(params) != self.nparams: raise Exception("params has not the correct number of parameters")
        parampos = 0
        for item in self._objs:
            item.setParams(params=params[parampos:parampos+item._nparams], priors=priors)
            parampos += item._nparams
        if self.oidata.systematic_fit:
            self.oidata.systematic_prior = params[parampos]
            self.oidata._systematic_prior = self.oidata.systematic_prior


    def compVis(self, params=None):
        """
        Calculate the complex visibility of the model from each separate object
        """
        if params is None: params = self.getP0() # initialize at p0 in case no params is given
        if self._tweakparams is not None: self._tweakparams(self, params)
        parampos = 0
        totflx = 0.
        totviscomp = _np.zeros(self.oidata.uvwl['u'].shape, dtype=complex) # initialize array
        for item in self._objs:
            viscomp, flx = item.compVis(oidata=self.oidata, params=params[parampos:parampos+item._nparams], flat=True)
            totviscomp += viscomp*flx
            totflx += flx
            parampos += item._nparams
        totviscomp /= totflx
        return self.oidata.remorph(totviscomp)

    def compuvimage(self, blmax, wl=None, params=None, nbpts=101):
        parampos = 0
        totFluxvis2 = 0.
        if wl is None: wl = self.oidata._wlmin+self.oidata._wlspan*0.5
        compvis = _np.zeros((nbpts, nbpts), dtype=complex) # initialize array
        u, v = _np.meshgrid(_np.linspace(-blmax, blmax, nbpts), _np.linspace(-blmax, blmax, nbpts))
        if self._tweakparams is not None: self._tweakparams(self, params)
        if params is not None: self.setParams(params)
        for item in self._objs:
            try:
                vis2 = item._calcCompVis(u=u, v=v, wl=wl, blwl=_np.hypot(u, v)/wl)
            except AttributeError:
                raise AttributeError("it looks like some parameters are not initialized. Input your parameters after params=[...]")
            compvis += vis2[0]*vis2[1]
            totFluxvis2 += vis2[1]
            parampos += item._nparams
        return compvis/totFluxvis2

    def compimage(self, params=None, sepmax=None, wl=None, masperpx=None, nbpts=101, psfConvolve=None, **kwargs):
        """
        psfConvolve in mas (lambda/D)
        """
        parampos = 0
        totFluxvis2 = 0.
        # check for set-resolution objects
        masperpxfixed = None
        nbptscheck = nbpts
        if nbpts is None: nbptscheck = int(2.*sepmax/masperpx)
        for item in self._objs:
            if hasattr(item, "_masperpx"):
                if masperpxfixed is None:
                    masperpxfixed = item._masperpx
                elif _np.abs(1-masperpxfixed/item._masperpx)*nbptscheck > 1:
                    raise Exception("can't return an image, it seems that several non-scalable images do not have the save masperpx")
        if masperpxfixed is not None: masperpx = masperpxfixed
        # check resolution parameters
        if sepmax is None and masperpx is None and nbpts is not None:
            sepmax = 1000*nbpts/(2*_np.hypot(self.oidata.uvwl['v'], self.oidata.uvwl['u']).max())
        if masperpx is not None and sepmax is not None:
            nbpts = int(2.*sepmax/masperpx)
        elif masperpx is not None and nbpts is not None:
            pass
        elif sepmax is not None and nbpts is not None:
            masperpx = sepmax*2./nbpts
        else:
            raise Exception("can't determine the resolution, nbpts should not be None")
        sepmax = 0.5*nbpts*masperpx
        # check other parameters
        if wl is None: wl = self.oidata.uvwl['wl'].min()
        # initialize array
        img = _np.zeros((nbpts, nbpts))
        if self._tweakparams is not None: self._tweakparams(self, params)
        if params is not None: self.setParams(params)
        for item in self._objs:
            try:
                img += item.image(sepmax=sepmax, masperpx=masperpx, wl=wl, nbpts=nbpts)
            except AttributeError:
                raise AttributeError("it looks like some parameters are not initialized. Input your parameters after params=[...]")
            parampos += item._nparams
        if psfConvolve is not None:
            if len(_core.aslist(psfConvolve))==1:
                psf = _core.psf(float(psfConvolve)*_core.MAS2RAD, masperpx)
            elif len(_core.aslist(psfConvolve))==3:
                sx, sy, th = map(float, psfConvolve[:3])
                y, x = _np.meshgrid(_np.arange(nbpts)-nbpts//2, _np.arange(nbpts)-nbpts//2)
                psf = _core.gauss2D(x, y, 1, 0, 0, sx*0.5/masperpx, sy*0.5/masperpx, th)
            img = _core.fftconvolve(img, psf, mode='same')
        if kwargs.get('retresol', False): return img, (sepmax, masperpx, nbpts)
        return img


    def image(self, params=None, sepmax=None, wl=None, masperpx=None, nbpts=101, cmap='jet', cm_min=None, cm_max=None, ret=False, visu=None, psfConvolve=None, **visuargs):
        """
        psfConvolve in mas (lambda/D)
        """
        # check other params
        if visu is None: visu = _core.ident
        if not callable(visu):
            print(_core.font.red+"ERROR: visu parameter must be callable"+_core.font.normal)
            return
        toplot, (sepmax, masperpx, nbpts) = self.compimage(sepmax=sepmax, wl=wl, params=params, nbpts=nbpts, psfConvolve=psfConvolve, retresol=True)
        toplot = visu(toplot, **visuargs)
        if cm_min is None: cm_min = toplot.min()
        if cm_max is None: cm_max = toplot.max()
        cmap, norm, mappable = _core.colorbar(cmap=cmap, cm_min=cm_min, cm_max=cm_max)
        thefig, ax = _core.astroskyplot(sepmax, polar=False, unit='mas')
        ax.matshow(toplot, origin='lower', extent=[-sepmax,sepmax,-sepmax,sepmax], cmap=cmap, norm=norm)
        _plt.colorbar(mappable)
        if ret: return toplot


    def uvimage(self, params=None, blmax=None, wl=None, typ='vis2', nbpts=101, cmap='jet', cm_min=None, cm_max=None, ret=False, visu=None, **visuargs):
        """
        typ can be: vis, vis2, phase
        """
        typdic = {'vis2':_core.abs2, 'phase':_np.angle, 'vis':_np.abs}
        # check other params
        if visu is None: visu = _core.ident
        if not callable(visu):
            print(_core.font.red+"ERROR: visu parameter must be callable"+_core.font.normal)
            return
        toplot = self.compuvimage(blmax=blmax, wl=wl, params=params, nbpts=nbpts)
        toplot = visu(typdic.get(typ, typdic['vis2'])(toplot), **visuargs)
        if cm_min is None: cm_min = toplot.min()
        if cm_max is None: cm_max = toplot.max()
        cmap, norm, mappable = _core.colorbar(cmap=cmap, cm_min=cm_min, cm_max=cm_max)
        thefig, ax = _core.astroskyplot(blmax, polar=False, unit='m')
        ax.matshow(toplot, origin='lower', extent=[-blmax,blmax,-blmax,blmax], cmap=cmap, norm=norm)
        _plt.colorbar(mappable)
        if ret: return toplot

    def residual(self, params, c=None, cmap='jet', cm_min=None, cm_max=None, datatype='All'):
        calcindex = {'vis2':0, 't3phi':1, 't3amp':2, 'visphi':3, 'visamp':4}
        fullmodel = self.compVis(params=params)
        cm_min_orig = cm_min
        cm_max_orig = cm_max
        if datatype.lower() == 'all':
            datatype = _core.DATAKEYSLOWER
        elif _np.iterable(datatype)==1 and not isinstance(datatype, str):
            datatype = [str(dum).lower() for dum in datatype if dum.lower() in _core.DATAKEYSLOWER]
        else:
            datatype = [str(dum).lower() for dum in [str(datatype)] if dum.lower() in _core.DATAKEYSLOWER]
        if self.oidata.systematic_fit: self.oidata.systematic_prior = params[self._nparamsObj]
        for index, datatypeloop in enumerate(datatype):
            if not getattr(self.oidata, datatypeloop): continue
            data = getattr(self.oidata, datatypeloop).data
            model = fullmodel[calcindex[datatypeloop]]
            error = getattr(self.oidata, datatypeloop).error
            if self.oidata.systematic_fit: error = _np.sqrt(error*error + self.oidata.systematic_prior*self.oidata.systematic_prior)
            thefig = _plt.figure()
            if c is not None:
                grid = _matplotlibGridspecGridSpec(5, 21)
            else:
                grid = _matplotlibGridspecGridSpec(5, 20)
            theax = thefig.add_subplot(grid[0:4,0:20])
            theotherax = thefig.add_subplot(grid[4,0:20], sharex=theax)
            for k, v in _np.ndenumerate(data):
                theax.add_patch(_matplotlibPatchesFancyArrowPatch((model[k], data[k]),(model[k], data[k]+error[k]), arrowstyle='-['))
                theax.add_patch(_matplotlibPatchesFancyArrowPatch((model[k], data[k]),(model[k], data[k]-error[k]), arrowstyle='-['))
            mini = min(data.min(), model.min())
            maxi = max(data.max(), model.max())
            theax.plot([mini, maxi], [mini, maxi], 'g-', lw=2)
            if c is not None:
                thecb = thefig.add_subplot(grid[:,20])
                thecb.set_xticklabels('')
                thecb.yaxis.tick_right()
                colorkeyattr = str(c)
                if not hasattr(getattr(self.oidata, datatypeloop), colorkeyattr): colorkeyattr = 'wl'
                colorattr = getattr(getattr(self.oidata, datatypeloop), colorkeyattr)
                #if datatypeloop[:2]=='t3':
                #    if colorattr.shape != self.oidata._t3outputshape:
                #        colorattr = colorattr.mean(axis=-1)
                if cm_min_orig is None: cm_min = colorattr.min()
                if cm_max_orig is None: cm_max = colorattr.max()
                cmap, norm, mappable = _core.colorbar(cmap=cmap, cm_min=cm_min, cm_max=cm_max)
                theax.scatter(x=model, y=data, s=30, c=colorattr, cmap=cmap, norm=norm, marker='o', edgecolors='none', alpha=1)
                div = _np.true_divide(model-data, error)
                theotherax.scatter(x=model, y=self.oidata.erb_sigma(_np.abs(div))*_np.sign(div), s=30, c=colorattr, cmap=cmap, norm=norm, marker='o', edgecolors='none', alpha=1)
                thefig.colorbar(mappable=mappable, cax=thecb)
                colortitle = ', color= '+colorkeyattr
            else:
                theax.scatter(x=model, y=data, s=30, marker='o', edgecolors='none', alpha=1)
                div = _np.true_divide(model-data, error)
                theotherax.scatter(x=model, y=self.oidata.erb_sigma(_np.abs(div))*_np.sign(div), s=30, marker='o', edgecolors='none', alpha=1)
                colortitle = ''
            theotherax.grid(True)
            theax.set_title(datatypeloop+colortitle)


    #def grid(self, param_x, param_y, bin_x=50, bin_y=50, mode=_np.max, cmap="jet"):
    #    pass

    def likelihood(self, params, customlike=None, chi2=False, **kwargs):
        kwargs['chi2'] = chi2
        return _likelihood(params=params, model=self, customlike=customlike, kwargs=kwargs)


    def save(self, filename, clobber=False):
        hdulist = _pf.HDUList()
        ext = '.oif.fits'
        if filename.find(ext)==-1: filename += ext
        hdu = _pf.PrimaryHDU()
        hdu.header.set('EXT', 'MODEL', comment='Type of information in the HDU')
        hdu.header.set('DATE', _strftime('%Y%m%dT%H%M%S'), comment='Creation Date')
        hdu.header.set('NOBJ', self.nobj, comment='Number of objects in the model')
        hdu.header.set('NPARAMS', self.nparams, comment='Total number of free parameters')
        for i, item in enumerate(self._objs):
            hdu.header.set('OBJ'+str(i), item.name, comment='Name of object '+str(i))
        allmodes = ['vis2', 't3phi', 't3amp', 'visphi', 'visamp']
        for mode in allmodes:
            number = []
            if getattr(self.oidata, mode):
                number = [len(i) for i in getattr(self.oidata, "_input_"+mode) if i is not None]
            hdu.header.set('N'+mode, int(_np.sum(number)), comment='Total number of '+mode+' measurements')
            hdu.header.set('F_'+mode, len(number), comment='Number of '+mode+' files')

        hdu.header.add_comment('Written by Guillaume SCHWORER')
        hdulist.append(hdu)

        hdulist.writeto(filename, clobber=clobber)

        for item in self._objs:
            item.save(filename, append=True, clobber=clobber)

        self.oidata.save(filename, append=True, clobber=clobber)

        return filename


def _likelihood(params, model, customlike=None, kwargs={}):
    '''
    Calculate the likelihood of the model versus the data.
    Use customlike and kwargs parameters for a custom likelihood function.
    This customlike function must accept input params:
    - params (list of the p values for that step)
    - modeledData (tuple of model vis2, t3phi, t3amp, visphi, visamp given the p value of the parameters). The uncalculated data is set to None.
    - model (oifiting object)

    And must return the log-likelihood of that step
    '''
    null = kwargs.pop('null', False)
    chi2 = kwargs.pop('chi2', False)
    quality = 0.
    ln_prior = 0.
    if model.oidata.systematic_fit:
        model.oidata.systematic_prior = params[model._nparamsObj]
    if null: # if we just want NULL hypothesis
        vis2, t3phi, t3amp, visphi, visamp = 1., 0., 1., 0., 1.
    else:
        parampos = 0
        for item in model._objs:
            for arg in item._pkeys:
                pb = getattr(item, arg+'_bounds')
                if not (pb[0] <= params[parampos] <= pb[1]):
                    return -_np.inf # exit with -inf if params outside bounds
                parampos += 1
        if model.oidata.systematic_fit:
            if not (model.oidata.systematic_bounds[0] <= model.oidata.systematic_prior <= model.oidata.systematic_bounds[1]):
                return -_np.inf # exit with -inf if params outside bounds
            #parampos += 1
        if customlike is not None: return customlike(params=params, model=self, **kwargs)
        if not chi2:
            pass
            #ln_prior += getattr(item, arg+'_prior_lnfunc')(params[parampos], params=params, parampos=parampos, **getattr(item, arg+'_prior_kwargs'))
        if model.nobj==0:
            vis2, t3phi, t3amp, visphi, visamp = 1., 0., 1., 0., 1.
        else:
            vis2, t3phi, t3amp, visphi, visamp = model.compVis(params=params)
    if model.oidata.vis2:
        if model.oidata.systematic_fit:
            invvar = 1./(model.oidata.vis2.error*model.oidata.vis2.error + model.oidata.systematic_prior*model.oidata.systematic_prior)
        else:
            invvar = model.oidata.vis2._invvar
        if chi2:
            quality += ((model.oidata.vis2.data - vis2)**2*invvar).sum()
        else:
            quality += ((model.oidata.vis2.data - vis2)**2*invvar - _np.log(invvar)).sum()
    if model.oidata.visphi:
        if model.oidata.systematic_fit:
            invvar = 1./(model.oidata.visphi.error*model.oidata.visphi.error + model.oidata.systematic_prior*model.oidata.systematic_prior)
        else:
            invvar = model.oidata.visphi._invvar
        if chi2:
            quality += ((model.oidata.visphi.data - visphi)**2*invvar).sum()
        else:
            quality += ((model.oidata.visphi.data - visphi)**2*invvar - _np.log(invvar)).sum()
    if model.oidata.visamp:
        if model.oidata.systematic_fit:
            invvar = 1./(model.oidata.visamp.error*model.oidata.visamp.error + model.oidata.systematic_prior*model.oidata.systematic_prior)
        else:
            invvar = model.oidata.visamp._invvar
        if chi2:
            quality += ((model.oidata.visamp.data - visamp)**2*invvar).sum()
        else:
            quality += ((model.oidata.visamp.data - visamp)**2*invvar - _np.log(invvar)).sum()
    if model.oidata.t3phi:
        if model.oidata.systematic_fit:
            invvar = 1./(model.oidata.t3phi.error*model.oidata.t3phi.error + model.oidata.systematic_prior*model.oidata.systematic_prior)
        else:
            invvar = model.oidata.t3phi._invvar
        if chi2:
            quality += ((model.oidata.t3phi.data - t3phi)**2*invvar).sum()
        else:
            quality += ((model.oidata.t3phi.data - t3phi)**2*invvar - _np.log(invvar)).sum()
    if model.oidata.t3amp:
        if model.oidata.systematic_fit:
            invvar = 1./(model.oidata.t3amp.error*model.oidata.t3amp.error + model.oidata.systematic_prior*model.oidata.systematic_prior)
        else:
            invvar = model.oidata.t3amp._invvar
        if chi2:
            quality += ((model.oidata.t3amp.data - t3amp)**2*invvar).sum()
        else:
            quality += ((model.oidata.t3amp.data - t3amp)**2*invvar - _np.log(invvar)).sum()
    if chi2:
        return quality
    else:
        return ln_prior - 0.5*quality # log like



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################


from scipy.special import j1 as airyJ1
from scipy.signal import fftconvolve
import numpy as np
from time import time as timetime
from multiprocessing import current_process as multiprocessingcurrent_process
from sys import maxsize as sysmaxint
sysmaxint = min(2**32, sysmaxint)
from scipy.interpolate import interp1d as scipyinterpolateinterp1d
from scipy.integrate import cumtrapz as scipyintegratecumtrapz
from matplotlib.cm import get_cmap as cmget_cmap
from matplotlib.cm import ScalarMappable as cmScalarMappable
from matplotlib.pyplot import Normalize as matplotlibpyplotNormalize
from matplotlib.pyplot import subplots as matplotlibpyplotsubplots


DATATYPEEXTNAMES = {'OI_T3':'T3', 'OI_VIS2':'VIS2', 'OI_VIS':'VIS'}
DATAEXTNAMES = {'OI_T3':'T3PHI', 'OI_VIS2':'VIS2DATA', 'OI_VIS':'VISPHI'}
ALLDATAEXTNAMES = {'OI_T3':['T3AMP','T3PHI'], 'OI_VIS2':['VIS2'], 'OI_VIS':['VISPHI','VISAMP']}
DATAKEYSDATATYPE = {'T3AMP':{'data':'T3AMP', 'error':'T3AMPERR', 'mask':'FLAG'},
                    'T3PHI':{'data':'T3PHI', 'error':'T3PHIERR', 'mask':'FLAG'},
                    'VIS2':{'data':'VIS2DATA', 'error':'VIS2ERR', 'mask':'FLAG'},
                    'VISPHI':{'data':'VISPHI', 'error':'VISPHIERR', 'mask':'FLAG'},
                    'VISAMP':{'data':'VISAMP', 'error':'VISAMPERR', 'mask':'FLAG'}
                    }
UVKEYSDATATYPE = {  'T3AMP':{'u1':'U1COORD', 'v1':'V1COORD', 'u2':'U2COORD', 'v2':'V2COORD'},
                    'T3PHI':{'u1':'U1COORD', 'v1':'V1COORD', 'u2':'U2COORD', 'v2':'V2COORD'},
                    'VIS2':{'u':'UCOORD', 'v':'VCOORD'},
                    'VISPHI':{'u':'UCOORD', 'v':'VCOORD'},
                    'VISAMP':{'u':'UCOORD', 'v':'VCOORD'}
                    }
ATTRDATATYPE = {'T3AMP':{'is_t3':True, 'is_angle':False},
                'T3PHI':{'is_t3':True, 'is_angle':True},
                'VIS2':{'is_t3':False, 'is_angle':False},
                'VISPHI':{'is_t3':False, 'is_angle':True},
                'VISAMP':{'is_t3':False, 'is_angle':False}
                }
DATAKEYSLOWER = ['vis2', 't3phi', 't3amp', 'visphi', 'visamp']

def abs2(ar): return np.abs(ar)**2


FCTVISCOMP = {'VIS2':abs2, 'T3PHI':np.angle, 'T3AMP':np.abs, 'VISPHI':np.angle, 'VISAMP':np.abs}

KEYSWL = {'wl':'EFF_WAVE', 'wl_d':'EFF_BAND'}
KEYSDATA = ['data', 'error', 'mask']
KEYSUV = ['v', 'u', 'wl', 'wl_d']
INPUTSAVEKEY = ['src', 'hduidx', 'hduwlidx', 'degree', 'flatten', 'significant_figures', 'indices', 'wlindices']


OLDNUMPY = int(np.__version__.split('.')[1])<10
MAS2RAD = 4.84813681109536e-09 # np.pi/1000./3600./180. # multiply mili-arc-second (mas) with that and you get radian
RAD2MAS = 1/MAS2RAD
ASEC2RAD = 4.84813681109536e-06 # np.pi/3600./180. # multiply arc second with that and you get radian
RAD2ASEC = 1/ASEC2RAD
DEG2RAD = 0.017453292519943295 # np.pi/180. # multiply degrees with that and you get radian
RAD2DEG = 1/DEG2RAD
DEUXPI = 2*np.pi
PISQRT2 = np.pi*np.sqrt(2)
LNDEUXPI = np.log(2*np.pi)
PISURDEUX = 0.5*np.pi
SECPERDAY = 86400.

def hduToDataType(hdu):
    """
    Give a hdu, return a datatype name as string or None if it is not a data hdu
    """
    return DATATYPEEXTNAMES.get(getattr(hdu, "header", {}).get('EXTNAME', None), None)

def hduWlindex(hdus, returnHdu=False):
    """
    Give a list of hdus, return the index of the OI_WAVELENGTH hdu, or False if not found
    """
    for idx, hdu in enumerate(hdus):
        if getattr(hdu, "header", {}).get('EXTNAME', None) == 'OI_WAVELENGTH':
            if returnHdu:
                return idx, hdu
            else:
                return idx
    else:
        return False

def hduToColNames(hdu):
    """
    Returns a list of the column names
    """
    return [i.name for i in hdu.data.columns.columns]

def replicate(ar, dims):
    left = ()
    right = ()
    hasbeennone = False
    for i in dims:
        if i is None or i==-1:
            hasbeennone = True
            continue
        if hasbeennone:
            right += (i,)
        else:
            left += (i,)
    ar = np.tile(ar, left+tuple(np.ones(ar.ndim).astype(int)))
    if right!=():
        ar = ar.repeat(np.prod(right)).reshape(ar.shape+right)
    return ar


class font:
    white = '\033[97m'
    black = '\033[38;5;16m'
    gray = '\033[90m'
    red = '\033[31m'
    green = '\033[32m'
    yellow = '\033[33m'
    orange = '\033[38;5;166m'
    blue = '\033[34m'
    magenta = '\033[35m'
    nocolor = '\033[39m'
    bold = '\033[1m'
    nobold = '\033[21m'
    underlined = '\033[4m'
    nounderlined = '\033[24m'
    dim = '\033[2m'
    nodim = '\033[22m'
    normal = nodim + nobold + nobold + nocolor
    clear = chr(27)+"[2J"
    def pos(self, line, col): return "\033["+str(int(line))+";"+str(int(col))+"H"


def astroskyplot(radius, polar=False, unit='arcsec', sameaxes=False):
    if polar:
        fig, ax = matplotlibpyplotsubplots(subplot_kw={'projection':'polar'})
        ax.set_theta_zero_location('N')
        ax.set_ylim([0, float(radius)])
        ax.set_xticklabels(['N',u'45°','E',u'135°','S',u'225°','W',u'315°'])
        ax.set_xlabel('['+unit+']')
    else:
        fig, ax = matplotlibpyplotsubplots()
        ax.set_xlim([-float(radius), float(radius)])
        ax.set_ylim([-float(radius), float(radius)])
        ax.invert_xaxis()
        ax.set_ylabel('E, dec ['+unit+']')
        ax.set_xlabel('S, ra ['+unit+']')
    if sameaxes: ax.set_aspect('equal')
    return fig, ax


def gen_seed(): return int(int(str((timetime()%3600)/3600)[2:])*multiprocessingcurrent_process().pid%sysmaxint)


def gen_generator(seed=None):
    if seed is None: return np.random.RandomState(gen_seed())
    return np.random.RandomState(seed)


def random_custom_pdf(x, pdf, size=1, renorm=False):
    if (np.asarray(pdf)<0).any(): raise Exception("Can't compute pdf with negative values") # if any negative value in pdf, raises error
    cdf = integrate_array(x, pdf)
    return random_custom_cdf(x, cdf, size=size, renorm=renorm)


def random_custom_cdf(x, cdf, size=1, renorm=False):
    cdf = np.array(cdf)
    if (np.diff(cdf)<0).any(): raise Exception("cdf must be monotonic increasing")
    if not renorm:
        if np.round(cdf[-1],10)!=1 or np.round(cdf[0],10)!=0: raise Exception("Wrong cdf distribution (first element not 0 or last element not 1)") # if first or last element not close enough from 0 and 1, rejects
    cdf = (cdf-cdf[0])/(cdf[-1]-cdf[0]) # forces first and last elements to be 0 and 1 exactly
    rnd = gen_generator()
    if size is False: return scipyinterpolateinterp1d(cdf, x, kind='linear')
    if size==1: return float(scipyinterpolateinterp1d(cdf, x, kind='linear')(rnd.uniform(size=1)))
    if np.size(size)==1: return scipyinterpolateinterp1d(cdf, x, kind='linear')(rnd.uniform(size=size))
    return scipyinterpolateinterp1d(cdf, x, kind='linear')(rnd.uniform(size=np.prod(size)).reshape(size))


def integrate_array(x, y, axis=0):
    y = np.array(y).swapaxes(axis,-1)
    inty = np.zeros(y.shape)
    myslice = [slice(0,i) for i in aslist(y.shape)[:-1]]+[slice(1,y.shape[-1])]
    inty[myslice] = scipyintegratecumtrapz(y, x)
    return inty.swapaxes(axis,-1)


def aslist(data, numpy=False, integer=False):
    if not hasattr(data, "__iter__") or isinstance(data, dict):
        if integer:
            ret = np.asarray([int(data)])
        else:
            ret = np.asarray([data])
    else:
        if integer:
            ret = np.asarray(data).flatten().astype(int)
        else:
            ret = np.asarray(data).flatten()
    if not numpy:
        return list(ret)
    return ret

def clean_name(txt):
    authorized=range(65,91)+range(48,58)+range(97,123)
    txt = str(txt)
    return "".join([(ord(letter) in authorized)*letter for letter in txt])


def colorbar(cmap="jet", cm_min=0, cm_max=1):
    if isinstance(cmap, str): cmap = cmget_cmap(cmap)
    norm = matplotlibpyplotNormalize(cm_min, cm_max)
    mappable = cmScalarMappable(cmap=cmap, norm=norm)
    mappable._A = []
    return cmap, norm, mappable

def ident(ar, *args, **kwargs): return ar

def _vv(img, masperpx, u, v, wl):
    demifov = 0.5*np.asarray(img.shape)*masperpx
    Y, X = np.meshgrid(np.linspace(demifov[0], -demifov[0], img.shape[0]), np.linspace(demifov[1], -demifov[1], img.shape[1]))
    X = X.reshape(X.shape+(1,))
    X = X.astype(np.float32)
    dum = X*u
    X = None
    Y = Y.reshape(Y.shape+(1,))
    Y = Y.astype(np.float32)
    dum += Y*v
    dum *= DEUXPI*MAS2RAD/wl
    return dum.astype(np.complex64)

def calcImgVis(img, masperpx, u, v, wl):
    """
    img: the image, (x, y) dimension
    masperpx the mas/px values for the images (x, y) in mas/px
    u, v: the coordinates on which to calculate the FT
    wl: the wavelength
    """
    dum = _vv(img, masperpx, u, v, wl)
    dum *= 1j
    dum = np.exp(dum)
    dum = dum.T
    dum *= img
    dum = dum.T
    if OLDNUMPY:
        return dum.sum(axis=0).sum(axis=0)
    else:
        return dum.sum(axis=(0,1))


def getDetails(hdu):
    allstation = hdu.data['STA_INDEX']
    ndata = allstation.shape[0]
    t3unique = 1.
    if allstation.shape[1] == 3:
        t3unique = np.exp(1./allstation[:,2])
    nunique = len(set((allstation[:,0] + 1j*allstation[:,1])*t3unique))
    nholes = np.arange(2, 100)
    if allstation.shape[1] == 3:
        nholes = nholes[np.argmin(np.abs(nholes*(nholes-1)*(nholes-2)/6-nunique))]
    else:
        nholes = nholes[np.argmin(np.abs(nholes*(nholes-1)/2-nunique))]
    nset = ndata/nunique
    if hdu.data[DATAEXTNAMES[hdu.header.get('EXTNAME')]].ndim == 1:
        nwl = 1
    else:
        nwl = hdu.data[DATAEXTNAMES[hdu.header.get('EXTNAME')]].shape[1]
    return ndata, nset, nunique, nholes, nwl

def unique(ar, precision, return_index=False, return_inverse=False, return_counts=False):
    """
    Returns the unique numbers of an array, at a certain precision

    Refer to np.unique
    """
    return np.unique(np.floor(ar/precision).astype(int)*precision, return_index=return_index, return_inverse=return_inverse, return_counts=return_counts)


def gethduMJD(hdu, withdet=False):
    ndata, nset, nunique, nholes, nwl = getDetails(hdu)
    mjd = hdu.data['MJD']
    if mjd.std() == 0.:
        delta = (hdu.data['INT_TIME'][np.arange(0, ndata, nunique)]*nunique)
        mjd = mjd + np.clip(delta, 30., 1e99).cumsum().repeat(nunique)/SECPERDAY
    sortmjd = np.argsort(mjd)
    if withdet:
        return np.arange(nset).repeat(nunique)[sortmjd], mjd, (ndata, nset, nunique, nholes, nwl)
    else:
        return np.arange(nset).repeat(nunique)[sortmjd], mjd


def maskedshape(shape, maskTrue):
    return "%s - %s%% masked" % (shape, np.round(maskTrue*1./(np.prod(shape) + maskTrue), 3)*100)


def round_fig(x, n=1, retint=False):
    """
    Rounds x at the n-th figure. n must be >1
    ex: 1234.567 with n=3-> 1230.0
    """
    if np.iterable(x):
        x = np.asarray(x).copy()
        ff = (x!=0)
        dd = 10**(np.floor(np.log10(np.abs(x[ff])))-n+1)
        x[ff] = np.round(x[ff]/dd)
        if not retint: x[ff] *= dd
        return x
    elif x != 0:
        dd = 10**(np.floor(np.log10(np.abs(x)))-n+1)
        x = np.round(x/dd)
        if not retint: x *= dd
        return x
    else:
        return x
    

def quantile(x, q, weights=None):
    """
    Like numpy.percentile, but:
    * Values of q are quantiles [0., 1.] rather than percentiles [0., 100.]
    * scalar q not supported (q must be iterable)
    * optional weights on x
    """
    if weights is None:
        return np.percentile(x, [100. * qi for qi in q])
    else:
        idx = np.argsort(x)
        xsorted = x[idx]
        cdf = np.add.accumulate(weights[idx])
        cdf /= cdf[-1]
        return np.interp(q, cdf, xsorted).tolist()

def psf(lOverd, masperpx):
    """
    lOverd in radian
    masperpx in mas per px
    """
    nbpts = (lOverd*4/(masperpx*MAS2RAD))//2*2+1
    y, x = np.meshgrid(np.linspace(-1,1,nbpts), np.linspace(-1,1,nbpts))
    psf = airy(np.hypot(y,x)*2*lOverd+1e-10, 1/lOverd)**2
    return psf/psf.sum()


def airy(th, B, lam=None):
    """
    Return the visibility value (unsquared), given
    - th the angular diameter in radian
    - B the baseline in m (alternatively in m/lambda)
    - lam the wavelength (alternatively None if B is already given as m/lambda)
    """
    if lam is None:
        x = np.pi*th*B
    else:
        x = np.pi*th*B/lam
    return 2*airyJ1(x)/x


def gauss2D(x, y, A, x0, y0, sigmaX, sigmaY, theta):
    sx2 = 1./sigmaX**2
    sy2 = 1./sigmaY**2
    st2 = np.sin(theta)**2
    ct2 = np.cos(theta)**2
    s2t = np.sin(2*theta)
    a = ct2*sx2 + st2*sy2
    b = s2t*(sy2 - sx2)
    c = st2*sx2 + ct2*sy2
    #a = (np.cos(theta)/sigmaX)**2 + (np.sin(theta)/sigmaY)**2
    #b = np.sin(2*theta)/sigmaY**2 - np.sin(2*theta)/sigmaX**2
    #c = (np.sin(theta)/sigmaX)**2 + (np.cos(theta)/sigmaY)**2
    return A*np.exp(-0.5*(a*(x-x0)**2 + b*(x-x0)*(y-y0) + c*(y-y0)**2))


def gauss1D(x, y, A, x0, y0, sigma):
    return A*np.exp(-0.5*((x-x0)**2 + (y-y0)**2)/sigma**2)

def gauss1Dsimple(blwlsigma):
    return np.exp(-blwlsigma*blwlsigma)


def BB(teff, wl, rel=False):
    """
    Give an effective temerature in K and a wavelength in m
    Returns the relative flux emitted (with respect to another temperature)
    """
    # pi*2*h*c**2 = 3.741771524664128e-16
    # h*c/kb = 0.014387769599838155
    return 1/(np.exp(0.014387769599838155/(wl*teff))-1)


class PriorWrapper(object):
    def __init__(self, fct, makenorm=True, makeunlog=False, fct_log=False, npts=1000, *args, **kwargs):
        if not callable(fct): raise Exception("Not callable")
        self.fct = fct
        self.args = args
        self.kwargs = kwargs
        self.makeunlog = makeunlog is True # apply exp on a ln-function at callback or not
        self.makenorm = makenorm is True # renorm at callback or not
        self.fct_log = fct_log is not False and makeunlog is False # whether to add or divide the normalization
        self.prior_log = kwargs.get('prior_log', False) # jeffrey
        if kwargs.get('prior_bounds', None) is None: raise Exception("No prior bounds found")
        if self.prior_log:
            self.x = np.logspace(np.log(kwargs.get('prior_bounds')[0]), np.log(kwargs.get('prior_bounds')[1]), npts, base=np.e)
        else:
            self.x = np.linspace(kwargs.get('prior_bounds')[0], kwargs.get('prior_bounds')[1], npts)
        if makenorm is True:
            if self.fct_log:
                self.normval = np.log(np.trapz(np.exp(self.fct(self.x, *self.args, **self.kwargs)), self.x))
            elif self.makeunlog:
                self.normval = np.trapz(np.exp(self.fct(self.x, *self.args, **self.kwargs)), self.x)
            else:
                self.normval = np.trapz(self.fct(self.x, *self.args, **self.kwargs), self.x)

    def __call__(self, x, *args, **kwargs):
        mergeargs = args + self.args[len(args):]
        mergekwargs = {}
        mergekwargs.update(self.kwargs, **kwargs)
        ret = self.fct(x, *mergeargs, **mergekwargs)
        if self.makeunlog: ret = np.exp(ret)
        if self.makenorm:
            if self.fct_log:
                ret -= self.normval
            else:
                ret /= self.normval
        return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################

import matplotlib.pyplot as _plt
try:
    from emcee import EnsembleSampler as _emceeEnsembleSampler
    _EMCEE = True
except ImportError:
    _EMCEE = False
from time import sleep as _sleep
from time import time as _time

from oimodel import _likelihood

import _core
_np = _core.np

from MCres import MCres as _MCres
from patiencebar import Patiencebar as _Patiencebar


class Oifiting(_MCres):
    def __init__(self, model, nwalkers=100, niters=500, burnInIts=100, threads=1, customlike=None, **kwargs):
        if 'sampler' in kwargs.keys():
            self._init(sampler=kwargs.pop('sampler'), paramstr=self.model.paramstr, nwalkers=self.nwalkers, niters=self.niters, burnInIts=self.burnInIts)
        elif _EMCEE:
            self.model = model
            self.nwalkers = int(nwalkers)
            self.niters = 0
            self.burnInIts = int(0 if burnInIts is None else burnInIts)
            self.sampler = _emceeEnsembleSampler(self.nwalkers, self.model.nparams, _likelihood, args=[self.model, customlike, kwargs], threads=max(1, int(threads)))
            self.run(niters=int(niters), burnInIts=self.burnInIts)

    def _init(self, sampler, paramstr=None, nwalkers=None, niters=None, burnInIts=None):
        super(Oifiting, self)._init(sampler=self.sampler, paramstr=self.model.paramstr, nwalkers=self.nwalkers, niters=self.niters, burnInIts=self.burnInIts)

    def _info(self):
        return _core.font.blue+"<SOIF Fit>%s\n %i walkers, %s burn-in, %s iters\n%s"%(_core.font.normal, self.nwalkers, self.burnInIts, self.niters, str(self.model))


    def run(self, niters, burnInIts=0):
        '''
        Start a MCMC simulation on as many CPUs as threads parameter. Use customlike parameter to use a custom likelihood function. This function must accept modeledData (ndarray) and model (Oifiting) as input parameters. It will also be given the kwargs that you give to this runmc function.
            e.g.: mycustomlike(modeledData, model, **kwargs)
        '''
        if not _EMCEE:
            print(_core.font.red+"Can't run more iterations, you don't have emcee installed"+_core.font.normal)
            return
        if self.model.nparams==0:
            print(_core.font.red+"ERROR: All parameters are set, there is no reason for MCMC"+_core.font.normal)
            return

        self.niters = getattr(self, "niters", 0) + int(niters)

        pb = _Patiencebar()
        print(_core.font.green+'Running emcee'+_core.font.normal)
        t0 = _time()

        if burnInIts>0:
            pb.reset(valmax=int(burnInIts), title="Burn in")
            p0 = [self.model.getP0() for i in range(self.nwalkers)]
            crackit = self.sampler.sample(p0, iterations=burnInIts, storechain=False)
            for res in crackit: pb.update()
        else: # no burn in
            res = [getattr(self, "_p0", None), None, getattr(self, "_rstate0", None)] # try init where we where
            if res[0] is None: res[0] = [self.model.getP0() for i in range(self.nwalkers)] # if first simu, no burn, was requested at all

        pb.reset(valmax=int(niters), title="Sampling")
        self.sampler.reset()

        crackit = self.sampler.sample(res[0], iterations=int(niters), rstate0=res[2])
        for res in crackit: pb.update()

        _sleep(0.1)
        print('Time elapsed = %.2fs' %(_time()-t0))
        print("Mean acceptance rate: %.3f" % (self.sampler.acceptance_fraction.mean()))

        self._p0 = res[0]
        self._rstate0 = res[2]
        self._init(sampler=self.sampler, paramstr=self.model.paramstr, nwalkers=self.nwalkers, niters=self.niters, burnInIts=self.burnInIts)


    def MCmap(self, param_x, param_y, bin_x=50, bin_y=50, cmap="jet", cm_min=None, cm_max=None, axescolor='w', polar=False, showmax=True, radec=False, **kwargs):
        """
        Return a 2D histogram of the MC chain, showing the walker density per bin
        """
        if radec:
            fig, ax = _core.astroskyplot(bin_y[-1], polar=polar, unit='mas')
        else:
            if polar:
                fig, ax = _plt.subplots(subplot_kw={'projection':'polar'})
                ax.set_theta_zero_location('N')
            else:
                fig, ax = _plt.subplots()
        self._map(param_x=param_x, param_y=param_y, fig=fig, ax=ax, bin_x=bin_x, bin_y=bin_y, cmap=cmap, cm_min=cm_min, cm_max=cm_max, axescolor=axescolor, polar=polar, showmax=showmax, **kwargs)


    def Pbmap(self, param_x, param_y, bin_x=50, bin_y=50, cmap="jet", cm_min=None, cm_max=None, axescolor='w', polar=False, showmax=True, radec=False, **kwargs):
        """
        Return a 2D histogram of the MC chain, showing the best loglikelihood per bin
        """
        if radec:
            fig, ax = _core.astroskyplot(bin_y[-1], polar=polar, unit='mas')
        else:
            if polar:
                fig, ax = _plt.subplots(subplot_kw={'projection':'polar'})
                ax.set_theta_zero_location('N')
            else:
                fig, ax = _plt.subplots()
        self._map(param_x=param_x, param_y=param_y, fig=fig, ax=ax, data=self.lnprob.compressed(), method=_np.max, bin_x=bin_x, bin_y=bin_y, cmap=cmap, cm_min=cm_min, cm_max=cm_max, axescolor=axescolor, polar=polar, showmax=showmax, **kwargs)


    def save(self, name, clobber=False):
        name = self.model.save(name, clobber=clobber)
        super(Oifiting, self).save(name=name, clobber=clobber, append=True)
        return name


    def statut(self, params=None, customlike=None, **kwargs):
        if params is None: params = self.best
        self.model.statut(params=params, customlike=customlike)


    def uvimage(self, blmax=None, wl=None, params=None, datatype='Data', nbpts=101, withdata=True, cmap='jet', cm_min=None, cm_max=None, ret=False):
        """
        Datatype in ['vis2', 'phase', 'vis']
        """
        allmodes = ['vis2', 'phase', 'vis']
        calctrick = {'vis2':_core.abs2, 'phase':_np.angle, 'vis':_np.abs}
        cm_min_orig = cm_min
        cm_max_orig = cm_max
        if datatype.lower() == 'all':
            datatype = allmodes
        if datatype.lower() == 'data':
            datatype = []
            if self.model.oidata.vis2: datatype.append('vis2')
            if self.model.oidata.visamp or self.model.oidata.t3amp: datatype.append('vis')
            if self.model.oidata.t3phi or self.model.oidata.visphi: datatype.append('phase')
        elif _np.iterable(datatype)==1 and not isinstance(datatype, str):
            datatype = [str(dum).lower() for dum in datatype if dum.lower() in allmodes]
        else:
            datatype = [str(dum).lower() for dum in [str(datatype)] if dum.lower() in allmodes]
        if blmax is None: blmax = _np.hypot(self.model.oidata.uvwl['v'], self.model.oidata.uvwl['u']).max()
        if wl is None: wl = self.model.oidata.uvwl['wl'].min()
        if params is None: params = self.best
        datashown = self.model.compuvimage(blmax=blmax, wl=wl, params=params, nbpts=nbpts)
        if ret: retval = []
        for index, datatypeloop in enumerate(datatype):
            data = calctrick[datatypeloop](datashown)
            if ret: retval.append(data)
            if cm_min_orig is None: cm_min = data.min()
            if cm_max_orig is None: cm_max = data.max()
            cmap, norm, mappable = _core.colorbar(cmap=cmap, cm_min=cm_min, cm_max=cm_max)
            thefig, ax = _core.astroskyplot(blmax, polar=False, unit='m')
            ax.matshow(data, origin='lower', extent=[-blmax,blmax,-blmax,blmax], cmap=cmap, norm=norm)
            if withdata:
                thefig.axes.scatter(getattr(self.model.oidata, datatypeloop).u, getattr(self.model.oidata, datatypeloop).v, c=getattr(self.model.oidata, datatypeloop).data, cmap=cmap, norm=norm)
            _plt.colorbar(mappable)
            thefig.axes.set_title(datatypeloop)
        if ret: return retval


    def image(self, params=None, sepmax=None, wl=None, masperpx=None, nbpts=101, cmap='jet', cm_min=None, cm_max=None, ret=False, visu=None, **visuargs):
        if params is None: params = self.best
        return self.model.image(params=params, sepmax=sepmax, wl=wl, masperpx=masperpx, nbpts=nbpts, cmap=cmap, cm_min=cm_min, cm_max=cm_max, ret=ret, visu=visu, **visuargs)


    def imagefft(self, sepmax=None, wl=None, params=None, nbpts=101, cmap='jet', cm_min=None, cm_max=None, ret=False, visu=None, **visuargs):
        if sepmax is None: sepmax = 1000*nbpts/(2*_np.hypot(self.model.oidata.uvwl['v'], self.model.oidata.uvwl['u']).max())
        if wl is None: wl = self.model.oidata.uvwl['wl'].min()
        if params is None: params = self.best
        if visu is None: visu = _np.abs
        #toplot = visu(self.model.compimage(sepmax=sepmax, wl=wl, params=params, nbpts=nbpts), **visuargs)
        toplot = _np.abs(_np.fft.fftshift(_np.fft.ifft2(self.model.compuvimage(blmax=nbpts*2*sepmax*_core.MAS2RAD/(_np.pi*wl), wl=wl, params=params, nbpts=nbpts))))
        if cm_min is None: cm_min = toplot.min()
        if cm_max is None: cm_max = toplot.max()
        cmap, norm, mappable = _core.colorbar(cmap=cmap, cm_min=cm_min, cm_max=cm_max)
        thefig, ax = _core.astroskyplot(sepmax, polar=False, unit='mas')
        ax.matshow(toplot, origin='lower', extent=[-sepmax,sepmax,-sepmax,sepmax], cmap=cmap, norm=norm)
        _plt.colorbar(mappable)
        if ret: return toplot


    def residual(self, params=None, c=None, cmap='jet', cm_min=None, cm_max=None, datatype='All'):
        if params is None: params = self.best
        self.model.residual(params=params, c=c, cmap=cmap, cm_min=cm_min, cm_max=cm_max, datatype=datatype)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################


from oiunitmodels import PointSource, UniformDisk, UniformDisk2D, UniformRing, UniformRing2D, Gauss, Gauss2D, BGimage, GaussDiff2D, GaussDiff, UniformDisk2DLinCR, UniformDiskLinCR



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from sys import argv, exit
import re, os

if "upl" in argv[1:]:
    import os
    os.system("python setup.py register -r pypi")
    os.system("python setup.py sdist upload -r pypi")
    exit()

try:
    from setuptools import setup
    setup
except ImportError:
    from distutils.core import setup
    setup

m = open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "soif", "_version.py")).read()
version = re.findall(r"__version__ *= *\"(.*?)\"", m)[0]


setup(
    name="soif",
    version=version,
    author="Guillaume Schworer",
    author_email="guillaume.schworer@obspm.fr",
    packages=["soif"],
    url="https://github.com/ceyzeriat/soif/",
    license="GNU General Public License v3 or later (GPLv3+)",
    description="Software for Optical Interferometry Fitting",
    long_description=open("README.rst").read() + "\n\n"
                    + "Changelog\n"
                    + "---------\n\n"
                    + open("HISTORY.rst").read(),
    package_data={"": ["LICENSE", "AUTHORS.rst"]},
    include_package_data=True,
    install_requires=["numpy","emcee","corner","patiencebar","MCres"],
    download_url = 'https://github.com/ceyzeriat/soif/tree/master/dist',
    keywords = ['astronomy','interferometry','data','processing','reduction','model','fitting','optical'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Astronomy"
    ],
)


# http://peterdowns.com/posts/first-time-with-pypi.html
